from collections import OrderedDict

from django.db.models import AutoField, BooleanField, CharField, DateField, \
    DateTimeField, DecimalField, FloatField, IntegerField, Model, \
    NullBooleanField, TextField


class DjangoQLSchema(object):
    exclude = ()  # models to exclude from introspection

    def __init__(self, model):
        assert issubclass(model, Model), 'Subclass of Django Model is expected'
        self.current_model = model
        self._models = None

    @property
    def models(self):
        if not self._models:
            self._models = self.introspect(
                model=self.current_model,
                exclude=tuple(self.model_label(m) for m in self.exclude),
            )
        return self._models

    def model_label(self, model):
        return str(model._meta)

    def introspect(self, model, exclude=()):
        """
        Start with given model and recursively walk through its relationships.
        
        Returns a dict with all model labels and their fields found.   
        """
        fields = OrderedDict()
        result = {self.model_label(model): fields}
        for field_name in self.get_fields(model):
            field = model._meta.get_field(field_name)
            if field.is_relation:
                if not field.related_model:
                    # GenericForeignKey
                    continue
                if field.related_model in self.exclude:
                    continue
                field_type = 'relation'
                relation = self.model_label(field.related_model)
                if relation not in exclude:
                    result.update(self.introspect(
                        model=field.related_model,
                        exclude=tuple(exclude) + tuple(result.keys()),
                    ))
            else:
                field_type = self.get_field_type(field)
                relation = None
            fields[field.name] = {
                'type': field_type,
                'relation': relation,
            }
        return result

    def get_fields(self, model):
        """
        By default, returns all field names of a given model. 
        
        Override this method to limit field options. You can either return a 
        plain list of field names from it, like ['id', 'name'], or call
        .super() and exclude unwanted fields from its result. 
        """
        return sorted(
            [f.name for f in model._meta.get_fields() if f.name != 'password']
        )

    def get_field_type(self, field):
        if isinstance(field, (AutoField, IntegerField)):
            return 'int'
        elif isinstance(field, (CharField, TextField)):
            return 'str'
        elif isinstance(field, (BooleanField, NullBooleanField)):
            return 'bool'
        elif isinstance(field, (DecimalField, FloatField)):
            return 'float'
        elif isinstance(field, DateTimeField):
            return 'datetime'
        elif isinstance(field, DateField):
            return 'date'
        return 'unknown'

    def as_dict(self):
        return {
            'current_model': self.model_label(self.current_model),
            'models': self.models,
        }



from django.db.models import Q, QuerySet

from .ast import Logical
from .parser import DjangoQLParser


def build_filter(expr):
    if isinstance(expr.operator, Logical):
        left, right = Q(build_filter(expr.left)), Q(build_filter(expr.right))
        if expr.operator.operator == 'or':
            return left | right
        else:
            return left & right
    search = '__'.join(expr.left.parts)
    invert = False
    op = {
        '=': '',
        '>': '__gt',
        '>=': '__gte',
        '<': '__lt',
        '<=': '__lte',
        '~': '__icontains',
        'in': '__in',
    }.get(expr.operator.operator)
    if op is None:
        op = {
            '!=': '',
            '!~': '__icontains',
            'not in': '__in',
        }[expr.operator.operator]
        invert = True
    q = Q(**{'%s%s' % (search, op): expr.right.value})
    return ~q if invert else q


def apply_search(queryset, search):
    """
    Applies search written in DjangoQL mini-language to given queryset
    """
    return queryset.filter(build_filter(DjangoQLParser().parse(search)))


class DjangoQLQuerySet(QuerySet):
    def djangoql(self, search):
        return apply_search(self, search)



import json

from django.conf.urls import url
from django.contrib import messages
from django.core.exceptions import FieldError, ValidationError
from django.http import HttpResponse
from django.views.generic import TemplateView

from .exceptions import DjangoQLSyntaxError
from .queryset import apply_search
from .schema import DjangoQLSchema


class DjangoQLSearchMixin(object):
    search_fields = ('_djangoql',)  # just a stub to have search input displayed
    djangoql_completion = True
    djangoql_schema = DjangoQLSchema
    djangoql_syntax_help_template = 'djangoql/syntax_help.html'

    def get_search_results(self, request, queryset, search_term):
        use_distinct = False
        if not search_term:
            return queryset, use_distinct
        try:
            return apply_search(queryset, search_term), use_distinct
        except (DjangoQLSyntaxError, ValueError, FieldError) as e:
            msg = str(e)
        except ValidationError as e:
            msg = e.messages[0]
        messages.add_message(request, messages.WARNING, msg)
        return queryset, use_distinct

    @property
    def media(self):
        media = super(DjangoQLSearchMixin, self).media
        if self.djangoql_completion:
            media.add_js((
                'djangoql/js/lib/lexer.js',
                'djangoql/js/completion.js',
                'djangoql/js/completion_admin.js',
            ))
            media.add_css({'': (
                'djangoql/css/completion.css',
                'djangoql/css/completion_admin.css',
            )})
        return media

    def get_urls(self):
        custom_urls = []
        if self.djangoql_completion:
            custom_urls += [
                url(
                    r'^introspect/$',
                    self.admin_site.admin_view(self.introspect),
                    name='%s_%s_djangoql_introspect' % (
                        self.model._meta.app_label,
                        self.model._meta.model_name,
                    ),
                ),
                url(
                    r'^djangoql-syntax/$',
                    TemplateView.as_view(
                        template_name=self.djangoql_syntax_help_template,
                    ),
                    name='djangoql_syntax_help',
                ),
            ]
        return custom_urls + super(DjangoQLSearchMixin, self).get_urls()

    def introspect(self, request):
        response = self.djangoql_schema(self.model).as_dict()
        return HttpResponse(
            content=json.dumps(response, indent=2),
            content_type='application/json; charset=utf-8',
        )
